@@ -27,15 +27,6 @@
 
 #include "verilated.h"  // for VerilatedMutex and clang annotations
 
-#ifndef VL_THREADED
-// Hitting this likely means verilated_threads.cpp is being compiled when
-// 'verilator --threads' was not used.  'verilator --threads' sets
-// VL_THREADED.
-// Alternatively it is always safe but may harm performance to always
-// define VL_THREADED for all compiles.
-#error "verilated_threads.h/cpp expected VL_THREADED (from verilator --threads)"
-#endif
-
 #include <atomic>
 #include <condition_variable>
 #include <set>
@@ -226,7 +217,7 @@ public:
     int numThreads() const { return m_workers.size(); }
     VlWorkerThread* workerp(int index) {
         assert(index >= 0);
-        assert(index < m_workers.size());
+        assert(static_cast<size_t>(index) < m_workers.size());
         return m_workers[index];
     }
 
